from .utils import RequestABC, to_prompt_structure, to_instruction, to_json_desc, find_json
from Agently.utils import RuntimeCtxNamespace
import httpx
import json

class Google(RequestABC):
    def __init__(self, request):
        self.request = request
        self.model_name = "Google"
        self.model_settings = RuntimeCtxNamespace(f"model.{ self.model_name }", self.request.settings)
        
    def construct_request_messages(self):
        #init request messages
        request_messages = []
        # - general instruction
        general_instruction_data = self.request.request_runtime_ctx.get_trace_back("prompt.general")
        if general_instruction_data:
            request_messages.append({ "role": "user", "parts": [{ "text": f"[YOUR GENERAL INSTRUCTION]:" }] })
            request_messages.append({ "role": "model", "parts": [{ "text": f"{ to_instruction(general_instruction_data) }" }] })
        # - role
        role_data = self.request.request_runtime_ctx.get_trace_back("prompt.role")
        if role_data:
            request_messages.append({ "role": "user", "parts": [{ "text": f"[YOUR ROLE DESCRIPTION]:" }] })
            request_messages.append({ "role": "model", "parts": [{ "text": f"{ to_instruction(role_data) }" }] })
        # - user info
        user_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.user_info")
        if user_info_data:
            request_messages.append({ "role": "user", "parts": [{ "text": f"[USER'S INTRODUCTION ABOUT HIMSELF/HERSELF]:" }] })
            request_messages.append({ "role": "model", "parts": [{ "text": f"{ to_instruction(user_info_data) }" }] })
        # - abstract
        headline_data = self.request.request_runtime_ctx.get_trace_back("prompt.abstract")
        if headline_data:
            request_messages.append({ "role": "user", "parts": [{ "text": f"[ABSTRACT ABOUT CHAT HISTORY BEFORE]:" }] })
            request_messages.append({ "role": "model", "parts": [{ "text": f"{ to_instruction(headline_data) }" }] })
        # - chat history
        chat_history_data = self.request.request_runtime_ctx.get_trace_back("prompt.chat_history")
        if chat_history_data:
            # make sure start with role "user"
            if chat_history_data[0]["role"] != "user":
                request_messages.append({ "role": "user", "parts": [{ "text": "[LATEST CHAT HISTORY]:" }] })
                current_role = "model"
            else:
                current_role = "user"
            for message in chat_history_data:
                if message["role"] != "user":
                    message["role"] = "model"
                if message["role"] == current_role:
                    request_messages.append({ "role": message["role"], "parts": [{ "text": message["content"] }] })
                else:
                    request_messages[-1]["parts"][0]["text"] += f"\n{ message['content'] }"
                current_role = "user" if current_role == "model" else "model"
            # make sure end with role "model"
            if request_messages[-1]["role"] != "model":
                request_messages.append({ "role": "model", "parts": [{ "text": "[CHAT HISTORY END]" }] })
        # - request message (prompt)
        prompt_input_data = self.request.request_runtime_ctx.get_trace_back("prompt.input")
        prompt_info_data = self.request.request_runtime_ctx.get_trace_back("prompt.info")
        prompt_instruct_data = self.request.request_runtime_ctx.get_trace_back("prompt.instruct")
        prompt_output_data = self.request.request_runtime_ctx.get_trace_back("prompt.output")
        # - files
        files_data = self.request.request_runtime_ctx.get_trace_back("prompt.files")
        # --- only input
        if not prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            raise Exception("[Request] Missing 'prompt.input', 'prompt.info', 'prompt.instruct', 'prompt.output' in request_runtime_ctx. At least set value to one of them.")
        prompt_text = ""
        if prompt_input_data and not prompt_info_data and not prompt_instruct_data and not prompt_output_data:
            prompt_text = to_instruction(prompt_input_data)
        # --- construct prompt
        else:
            prompt_dict = {}
            if prompt_input_data:
                prompt_dict["[INPUT]"] = to_instruction(prompt_input_data)
            if prompt_info_data:
                prompt_dict["[HELPFUL INFORMATION]"] = str(prompt_info_data)
            if prompt_instruct_data:
                prompt_dict["[INSTRUCTION]"] = to_instruction(prompt_instruct_data)
            if prompt_output_data:
                if isinstance(prompt_output_data, (dict, list, set)):
                    prompt_dict["[OUTPUT REQUIREMENT]"] = {
                        "TYPE": "JSON can be parsed in runtime",
                        "FORMAT": to_json_desc(prompt_output_data),
                    }
                    self.request.request_runtime_ctx.set("response:type", "JSON")
                else:
                    prompt_dict["[OUTPUT REQUIERMENT]"] = str(prompt_output_data)
            prompt_text = to_prompt_structure(prompt_dict, end="[OUTPUT]:\n")
        request_messages.append({ "role": "user", "parts": [{ "text": prompt_text }] })
        return request_messages

    def generate_request_data(self):
        options = self.model_settings.get_trace_back("options", {})
        return {
            "messages": self.construct_request_messages(),
            "options": options,
        }

    async def request_model(self, request_data: dict):
        api_key = self.model_settings.get_trace_back("auth.api_key")
        proxy = self.request.settings.get_trace_back("proxy")
        messages = request_data["messages"]
        options = request_data["options"]
        request_params = { 
            "data": json.dumps({ "contents": messages, **options }),
            "timeout": None,
        }
        client_params = {}
        if proxy:
            client_params["proxy"] = proxy
        async with httpx.AsyncClient(**client_params) as client:
            async with client.stream(
                "POST",
                f"https://generativelanguage.googleapis.com/v1/models/gemini-pro:streamGenerateContent?key={ api_key }",
                **request_params
            ) as response:
                async for chunk in response.aiter_lines():
                    yield chunk

    async def broadcast_response(self, response_generator):
        full_content = ""
        buffer = ""
        async for chunk in response_generator:
            full_content += chunk
            if '"text": "' in chunk:
                content = json.loads(f"{{ {chunk} }}")
                yield({ "event": "response:delta", "data": content["text"] })
                buffer += content["text"]
        if buffer == "":
            raise Exception(f"[Request]Google Gemini Error: { full_content }")
        yield({ "event": "response:done_origin", "data": json.loads(full_content) })
        yield({ "event": "response:done", "data": buffer })

    def export(self):
        return {
            "generate_request_data": self.generate_request_data,
            "request_model": self.request_model,
            "broadcast_response": self.broadcast_response,
        }

def export():
    return ("Google", Google)


import types
import asyncio
from .utils import ComponentABC
from Agently.utils import RuntimeCtx, RuntimeCtxNamespace

class EventListener(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.agent_listeners = RuntimeCtxNamespace("event_listeners", self.agent.agent_runtime_ctx)
        self.listeners = RuntimeCtxNamespace("event_listeners", RuntimeCtx(parent=self.agent.agent_runtime_ctx))
        self.listeners.set({})
        self.async_tasks = []

    def add(self, event:str, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        if is_agent_event:
            if event not in (self.agent_listeners.get(trace_back=False) or {}):
                self.agent_listeners.update(event, [])
            self.agent_listeners.append(event, { "listener": listener, "is_await": is_await })
            return self.agent
        else:
            if event not in (self.listeners.get(trace_back=False) or {}):
                self.listeners.update(event, [])
            self.listeners.append(event, { "listener": listener, "is_await": is_await })
            return self.agent

    def on_delta(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:delta", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_done(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:done", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent

    def on_finally(self, listener: callable, *, is_await:bool=False, is_agent_event:bool=False):
        self.add("response:finally", listener, is_await=is_await, is_agent_event=is_agent_event)
        return self.agent
    
    async def call_event_listeners(self, event: str, data: any):
        listeners = self.listeners.get_trace_back() or {}
        if event in listeners:
            for listener_info in listeners[event]:
                if asyncio.iscoroutinefunction(listener_info["listener"]):
                    if listener_info["is_await"]:
                        await listener_info["listener"](data)
                    else:
                        self.async_tasks.append(asyncio.create_task(listener_info["listener"](data)))
                else:
                    listener_info["listener"](data)
        return self.agent

    async def _suffix(self, event:str, data: any):
        await self.call_event_listeners(event, data)
        if event == "response:finally":
            for async_task in self.async_tasks:
                await async_task
            self.async_tasks = []
            self.listeners.set({})

    def export(self):
        return {
            "prefix": None,
            "suffix": self._suffix,
            "alias": {
                "add_event_listener": { "func": self.add },
                "on_delta": { "func": self.on_delta },
                "on_done": { "func": self.on_done },
                "on_finally": { "func": self.on_finally },
                "call_event_listeners": { "func": self.call_event_listeners },
            },
        }

def export():
    return ("EventListener", EventListener)


import json5
import asyncio
import threading
import queue
from ..Request import Request
from ..utils import RuntimeCtx, StorageDelegate, PluginManager, AliasManager, ToolManager, IdGenerator, DataGenerator,check_version, load_json

class Agent(object):
    def __init__(
        self,
        *,
        agent_id: str=None,
        auto_save: bool=False,
        parent_agent_runtime_ctx: object,
        parent_tool_manager: object,
        global_storage: object,
        global_websocket_server: object,
        parent_plugin_manager: object,
        parent_settings: object,
        is_debug: bool=False
    ):
        # Integrate
        self.global_storage = global_storage
        self.global_websocket_server = global_websocket_server
        self.plugin_manager = PluginManager(parent = parent_plugin_manager)
        self.tool_manager = ToolManager(parent = parent_tool_manager)
        self.settings = RuntimeCtx(parent = parent_settings)
        self.alias_manager = AliasManager(self)
        self.agent_runtime_ctx = RuntimeCtx(parent = parent_agent_runtime_ctx)
        self.request_runtime_ctx = RuntimeCtx()
        self.request = Request(
            parent_plugin_manager = self.plugin_manager,
            parent_request_runtime_ctx = self.request_runtime_ctx,
            parent_settings = self.settings,
        )
        self.worker_request = Request(
            parent_plugin_manager = self.plugin_manager,
            parent_request_runtime_ctx = None,
            parent_settings = self.settings,
        )
        # Debug
        self.settings.set("is_debug", is_debug)
        # Agent Id
        if agent_id == None or agent_id == "":
            self.agent_id = IdGenerator("agent").create()
        else:
            self.agent_id = agent_id
        # Agent Storage
        self.agent_storage = StorageDelegate(
            db_name = self.agent_id,
            plugin_manager = self.plugin_manager,
            settings = self.settings,
        )
        # Load Saved agent_runtime_ctx
        self.agent_runtime_ctx.update_by_dict(self.agent_storage.table("agent_runtime_ctx").get())
        # Set Agent Auto Save Setting
        self.agent_runtime_ctx.set("agent_auto_save", auto_save)
        # Version Check In Debug Model
        """
        if self.settings.get_trace_back("is_debug"):
            check_version_record = self.global_storage.get("agently", "check_version_record")
            today = str(datetime.date.today())
            if check_version_record != today:
                check_version(self.global_storage, today)
        """
        # Agent Request Prefix & Suffix
        self.agent_request_prefix = {}
        self.agent_request_suffix = []
        # Agent Response Generator
        self.response_generator = DataGenerator()
        # Register Default Request Alias to Agent
        self.request._register_default_alias(self.alias_manager)
        # Install Agent Components
        self.refresh_plugins()

    def refresh_plugins(self):
        self.alias_manager.empty_alias()
        # Agent Components
        agent_components = self.plugin_manager.get("agent_component")
        component_toggles = self.settings.get_trace_back("component_toggles")
        for agent_component_name, AgentComponentClass in agent_components.items():
            # Skip component_toggles Those Be Toggled Off
            if agent_component_name in component_toggles and component_toggles[agent_component_name] == False:
                continue
            # Attach Component
            agent_component_instance = AgentComponentClass(agent = self)
            setattr(self, agent_component_name, agent_component_instance)
            component_export = agent_component_instance.export()
            # Register export_prefix, export_suffix
            if "prefix" in component_export:
                if callable(component_export["prefix"]):
                    component_export["prefix"] = [component_export["prefix"]]
                self.agent_request_prefix.update({ agent_component_name: component_export["prefix"] })
            if "suffix" in component_export:
                if isinstance(component_export["suffix"], list):
                    self.agent_request_suffix.extend(component_export["suffix"])
                elif callable(component_export["suffix"]):
                    self.agent_request_suffix.append(component_export["suffix"])
            # Register Alias
            if "alias" in component_export:
                for alias_name, alias_info in component_export["alias"].items():
                    self.alias_manager.register(
                        alias_name,
                        alias_info["func"],
                        return_value = alias_info["return_value"] if "return_value" in alias_info else False,
                        agent_component_name = agent_component_name,
                    )

    def toggle_auto_save(self, is_enabled: bool):
        self.agent_runtime_ctx.set("agent_auto_save", is_enabled)
        return self

    def toggle_component(self, component_name: str, is_enabled: bool):
        self.settings.set(f"component_toggles.{ component_name }", is_enabled)
        self.refresh_plugins()
        return self

    def save(self):
        self.agent_storage.table("agent_runtime_ctx").update_by_dict(self.agent_runtime_ctx.get()).save()
        return self

    def set_settings(self, settings_key: str, settings_value: any):
        self.settings.set(settings_key, settings_value)
        return self

    def set_global_variable(self, variable_name: str, variable_value: any):
        self.settings.set(f"global_variables.{ variable_name }", variable_value)
        return self

    def set_agent_prompt(self, key: str, value: any):
        self.agent_runtime_ctx.set(f"prompt.{ key }", value)
        return self

    def get_agent_prompt(self, key: str):
        return self.agent_runtime_ctx.get(f"prompt.{ key }")

    def remove_agent_prompt(self, key: str):
        self.agent_runtime_ctx.remove(f"prompt.{ key }")
        return self

    async def start_async(self, request_type: str=None, *, return_generator:bool=False):
        try:
            is_debug = self.settings.get_trace_back("is_debug")
            # Auto Save Agent runtime_ctx
            if self.agent_runtime_ctx.get("agent_auto_save") ==  True:
                self.save()
            # Load Prompt in Agent runtime_ctx to Request runtime_ctx
            self.request_runtime_ctx.set("prompt", self.agent_runtime_ctx.get("prompt", {}))
            # Call Prefix Funcs to Prepare Prefix Data(From agent_runtime_ctx To request_runtime_ctx)
            async def call_prefix_funcs(prefix_funcs):
                if prefix_funcs != None:
                    for prefix_func in prefix_funcs:
                        prefix_data = await prefix_func() if asyncio.iscoroutinefunction(prefix_func) else prefix_func()
                        if prefix_data != None:
                            if isinstance(prefix_data, tuple) and isinstance(prefix_data[0], str) and prefix_data[1] != None:
                                key = prefix_data[0] if prefix_data[0].startswith("prompt.") else f"prompt.{ prefix_data[0] }"
                                self.request.request_runtime_ctx.update(key, prefix_data[1])
                            elif isinstance(prefix_data, dict):
                                for key, value in prefix_data.items():
                                    key = key if key.startswith("prompt.") else f"prompt.{ key }"
                                    if value != None:
                                        self.request.request_runtime_ctx.delta(key, value)
                            else:
                                raise Exception("[Agent Component] Prefix return data error: only accept None or Dict({'<request slot name>': <data append to slot>, ... } or Tuple('request slot name', <data append to slot>)")

            prefix_orders = self.settings.get_trace_back("plugin_settings.agent_component.orders.prefix")
            for agent_component_name in prefix_orders["firstly"]:
                if agent_component_name in self.agent_request_prefix:
                    await call_prefix_funcs(self.agent_request_prefix[agent_component_name])
            for agent_component_name in prefix_orders["normally"]:
                if agent_component_name in self.agent_request_prefix:
                    await call_prefix_funcs(self.agent_request_prefix[agent_component_name])
            for agent_component_name in prefix_orders["finally"]:
                if agent_component_name in self.agent_request_prefix:
                    await call_prefix_funcs(self.agent_request_prefix[agent_component_name])

            # Request
            event_generator = await self.request.get_event_generator(request_type)
        
            # Call Suffix Func to Handle Response Events
            if is_debug:
                print("[Realtime Response]\n")
            async def call_request_suffix(response):
                for suffix_func in self.agent_request_suffix:
                    if asyncio.iscoroutinefunction(suffix_func):
                        await suffix_func(response["event"], response["data"]) 
                    else:
                        suffix_func(response["event"], response["data"])

            async def handle_response(response):
                if response["event"] == "response:delta":
                    if is_debug:
                        print(response["data"], end="")
                    self.response_generator.add(response["data"])
                if response["event"] == "response:done":
                    if self.request.response_cache["reply"] == None:
                        self.request.response_cache["reply"] = response["data"]
                    if is_debug:
                        print("\n--------------------------\n")
                        print("[Final Reply]\n", self.request.response_cache["reply"], "\n--------------------------\n")
                    self.response_generator.end()
                await call_request_suffix(response)

            await handle_response({ "event": "response:start", "data": {} })
            if "__aiter__" in dir(event_generator):
                async for response in event_generator:
                    await handle_response(response)
            else:
                for response in event_generator:
                    await handle_response(response)
            
            # Load JSON and fix if Required
            if (
                isinstance(self.request.response_cache["prompt"]["output"], (dict, list, set))
                or self.request.response_cache["type"] == "JSON"
            ):
                temp_request = (
                    Request()
                        .set_settings("current_model", self.request.settings.get_trace_back("current_model"))
                        .set_settings("model", self.request.settings.get_trace_back("model"))
                )
                reply = await load_json(
                    self.request.response_cache["reply"],
                    self.request.response_cache["prompt"]["input"],
                    self.request.response_cache["prompt"]["output"],
                    temp_request,
                    is_debug = is_debug,
                )
                if isinstance(reply, str) and reply.startswith("$$$JSON_ERROR:"):
                    raise Exception(reply[14:])
                self.request.response_cache["reply"] = json5.loads(reply) if isinstance(reply, str) else reply

            await handle_response({ "event": "response:finally", "data": self.request.response_cache })

            if return_generator:
                self.request_runtime_ctx.empty()
                return self.response_generator.start()
            else:
                self.request_runtime_ctx.empty()
                return self.request.response_cache["reply"]
        except Exception as e:
            self.response_generator.end()
            self.request_runtime_ctx.empty()
            raise(e)

    def start(self, request_type: str=None, *, return_generator:bool=False):
        reply_queue = queue.Queue()
        is_debug = self.settings.get_trace_back("is_debug")
        def start_in_theard():
            asyncio.set_event_loop(asyncio.new_event_loop())
            loop = asyncio.get_event_loop()
            if is_debug:
                reply = loop.run_until_complete(self.start_async(request_type))
                reply_queue.put_nowait(reply)
            else:
                try:
                    reply = loop.run_until_complete(self.start_async(request_type))
                    reply_queue.put_nowait(reply)
                except Exception as e:
                    reply = None
                    raise Exception(f"[Agent Request] Error: { str(e) }")                    
                finally:
                    loop.close()
        theard = threading.Thread(target=start_in_theard)
        theard.start()
        if return_generator:
            return self.response_generator.start()
        else:
            theard.join()
            try:        
                reply = reply_queue.get_nowait()
            except:
                reply = None
            return reply

    def start_websocket_server(self, port:int=15365):
        is_debug = self.settings.get_trace_back("is_debug")

        def alias_handler(data: any, response: callable):
            try:
                if isinstance(data["params"], dict):
                    getattr(self, data["name"])(**data["params"])
                elif isinstance(data["params"], (list, tuple, set)): 
                    getattr(self, data["name"])(*data["params"])
                else:
                    getattr(self, data["name"])(data["params"])
                response("alias_done")
            except Exception as e:
                if is_debug:
                    print("[Agent WebSocket Server] Error: ", str(e))
        self.global_websocket_server.add_event_handler(self.agent_id, "alias", alias_handler)

        def start_handler(data: any, response: callable):
            def write_message(event: str, data: any):
                try:
                    response(event, data)
                except Exception as e:
                    if is_debug:
                        print("[Agent WebSocket Server] Error", str(e))
            self\
                .on_delta(
                    lambda data: write_message("delta", data)
                )\
                .on_done(
                    lambda data: write_message("done", data)
                )\
                .start()
        self.global_websocket_server.add_event_handler(self.agent_id, "start", start_handler)

        if is_debug:
            print(f"[WebSocket Server] Event listeners of agent '{ self.agent_id }' are on.")

        if self.global_websocket_server.status == 0:
            self.global_websocket_server.start(port)
            if is_debug:
                print(f"[WebSocket Server] WebSocket server started at { self.global_websocket_server.host }:{ self.global_websocket_server.port }.")
        else:
            if is_debug and port != self.global_websocket_server.port:
                print(f"[WebSocket Server] WebSocket server has already started at { self.global_websocket_server.host }:{ self.global_websocket_server.port }")

    def stop_websocket_server(self):
        self.global_websocket_server.remove_event_handler(self.agent_id)
        if self.settings.get_trace_back("is_debug"):
            print(f"[WebSocket Server] Agent '{ self.agent_id } websocket server stoped.'")


import asyncio
from .utils import ComponentABC
from Agently.utils import RuntimeCtxNamespace
from openai import OpenAI
import httpx

class OpenAIAssistant(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent
        self.assistant_runtime_ctx = RuntimeCtxNamespace("openai_assistant", self.agent.agent_runtime_ctx)

    def use_assistant(self, use_assistant: bool):
        self.agent.request.set_model("use_assistant", use_assistant)
        return self

    def set_assistant_id(self, assistant_id: str):
        self.agent.request.set_model("assistant_id", assistant_id)
        self.agent.request.set_model("use_assistant", True)
        return self

    def set_name(self, name: str):
        self.assistant_runtime_ctx.set("name", name)
        return self

    def set_instruction(self, instruction: str):
        self.assistant_runtime_ctx.set("instruction", instruction)
        return self

    def set_model(self, model: str):
        self.agent.request.set_model("options.model", model)
        return self

    def set_api_key(self, api_key: str):
        self.agent.request.set_model_auth({ "api_key": api_key })
        return self

    def set_proxy(self, proxy: str):
        self.agent.request.set_proxy(proxy)
        return self

    def append_file(self, file_name: str, file_path: str):
        self.assistant_runtime_ctx.update(f"files.{ file_name }.path", file_path)
        return self

    def get_file_id(self, file_name: str):
        return self.assistant_runtime_ctx.get(f"files.{filename}.id")

    def remove_file(self, file_name: str):
        self.assistant_runtime_ctx.remove(f"files.{ file_name }")
        return self

    def append_tool(self, tool_type: str, **kwargs):
        if { "type": tool_type, **kwargs } not in self.assistant_runtime_ctx.get("tools", []):
            self.assistant_runtime_ctx.append("tools", { "type": tool_type, **kwargs })
        return self

    def use_code_interpreter(self):
        self.append_tool("code_interpreter")
        return self

    def append_function(self, function_info: dict):
        self.append_tool(self, "function", function = function_info)
        return self

    def _create_client(self):
        client_params = {}
        base_url = self.agent.request.settings.get_trace_back("model.OpenAI.url")
        if base_url:
            client_params.update({ "base_url": base_url })
        proxy = self.agent.request.settings.get_trace_back("proxy")
        if proxy:
            client_params.update({ "http_client": httpx.Client( proxies = proxy ) })
        api_key = self.agent.request.settings.get_trace_back("model.OpenAI.auth.api_key")
        if api_key:
            client_params.update({ "api_key": api_key })
        else:
            raise Exception("[Request] OpenAI require api_key. use .set_auth({ 'api_key': '<Your-API-Key>' }) to set it.")
        client = OpenAI(**client_params)
        return client

    def update(self, *, force_to_create: bool=False):
        # create client
        client = self._create_client()
        # check file change
        files = self.assistant_runtime_ctx.get("files", {})
        file_ids = []
        for file_name, file_info in files.items():
            if "id" not in file_info:
                file = client.files.create(
                    file=open(file_info["path"], "rb"),
                    purpose="assistants",
                )
                file_info["id"] = file.id
                self.assistant_runtime_ctx.set(f"files.{ file_name }.id", file_info["id"])
            file_ids.append(file_info["id"])
        # prepare assisant info
        name = self.assistant_runtime_ctx.get("name", "My Assistant")
        instruction = self.assistant_runtime_ctx.get("instruction", "You are a helpful assistant.")
        model = self.agent.request.settings.get_trace_back(
            "model.OpenAI.options.model",
            "gpt-3.5-turbo-1106",
        )
        tools = self.assistant_runtime_ctx.get("tools", [])
        
        assistant_id = self.agent.request.settings.get_trace_back("model.OpenAI.assistant_id")
        # update
        if assistant_id and not force_to_create:
            assistant = client.beta.assistants.update(
                assistant_id = assistant_id,
                name = name,
                instructions = instruction,
                tools = tools,
                model = model,
                file_ids = file_ids
            )
        # create
        else:
            assistant = client.beta.assistants.create(
                name = name,
                instructions = instruction,
                tools = tools,
                model = model,
                file_ids = file_ids,
            )
            assistant_id = assistant.id
            self.set_assistant_id(assistant.id)
        # turn on setting 'use_assistant'
        self.agent.request.set_model("use_assistant", True)
        # and save agent to keep all information
        self.agent.save()
        return assistant_id

    def create(self):
        return self.update(force_to_create=True)

    def export(self):
        return {
            "prefix": None,
            "suffix": None,
            "alias": {},
        }

def export():
    return ("OpenAIAssistant", OpenAIAssistant)


import queue

class DataGeneratorTimeoutException(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"\n[Data Generator] Error: { str(self.message) }"

class DataGeneratorEvent:
    def __init__(self, data_generator: "DataGenerator"):
        self.data_generator = data_generator
        self.handlers = {}
    
    def emit(self, event:str, data:any):
        self.data_generator.add({
            "event": event,
            "data": data,
        })
        return self
    
    def on(self, event:str, handler:callable):
        if event not in self.handlers:
            self.handlers.update({ event: [] })
        self.handlers[event].append(handler)
        return self
    
    def clean(self, event:str):
        if event in self.handlers:
            del self.handlers[event]
        return self

class DataGenerator:
    def __init__(self, *, timeout:int=None, allow_timeout:bool=False):
        self.queue = queue.Queue()
        self.timeout = timeout
        self.allow_timeout = allow_timeout
        self.event = DataGeneratorEvent(self)
    
    def start(self):
        end_flag = False
        while True:
            try:
                if self.timeout:
                    data = self.queue.get(timeout=self.timeout)
                else:
                    data = self.queue.get()
                if data == "$END$":
                    self.queue = queue.Queue()
                    break
                else:
                    if "event" in data:
                        if data["event"] in self.event.handlers:
                            for handler in self.event.handlers[data["event"]]:
                                result = handler(data["data"])
                                if "yield" in result:
                                    yield result["yield"]
                                if "end" in result and result["end"] == True:
                                    self.queue = queue.Queue()
                                    end_flag = True
                            if end_flag:
                                break
                    else:
                        yield data
            except queue.Empty:
                self.queue = queue.Queue()
                if self.allow_timeout:
                    break
                else:
                    raise DataGeneratorTimeoutException(f"Data generator timeout (wait { self.timeout } seconds). Use `.end()` to put an end to the data queue before `.start()`.")
    
    def add(self, data):
        self.queue.put(data)
        return self
        
    def end(self):
        self.queue.put("$END$")


import copy
import logging
from .utils.verify import validate_dict
from .utils.find import has_target_by_attr, find_by_attr, find
from .utils.chunk_helper import create_connection_symbol
from .Chunk import SchemaChunk
from .lib.constants import DEFAULT_INPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE_VALUE, DEFAULT_OUTPUT_HANDLE, DEFAULT_INPUT_HANDLE, EXECUTOR_TYPE_NORMAL

class Schema:
    """Workflow 的描述"""

    def __init__(self, schema_data = { 'chunks': [], 'edges': [] }, workflow = None, logger: logging.Logger = None):
        self.logger = logger
        self._chunks = []
        self._edges = []
        # 依次调用添加方法添加
        (
            self
                .append_raw_chunk_list(schema_data.get('chunks', []))
                .connect_with_edges(schema_data.get('edges', []))
        )
        self.workflow = workflow
    
    def compile(self):
        """编译处理，得到可用数据"""
        compiled_chunks = copy.deepcopy(self._chunks)
        compiled_edges = copy.deepcopy(self._edges)
        # 处理 1：收集 handle 使用，自动完成 handle 注册
        # 构建 chunk map
        chunk_map = {}
        for chunk in compiled_chunks:
            chunk_map[chunk['id']] = chunk
        # 从连接关系中搜集使用到的 Handle，尝试自动帮助注册
        for edge in compiled_edges:
            source_chunk = chunk_map.get(edge['source'])
            target_chunk = chunk_map.get(edge['target'])
            # 合法校验
            if not source_chunk:
                raise ValueError(
                    f"Compilation failed due to an undefined chunk: '{edge['source']}'")
            elif not target_chunk:
                raise ValueError(
                    f"Compilation failed due to an undefined chunk: '{edge['target']}'")
            # 自动注册使用到的 handle
            if not has_target_by_attr(source_chunk['handles']['outputs'], 'handle', edge['source_handle']):
                source_chunk['handles']['outputs'].append({
                    'handle': edge['source_handle']
                })
            if not has_target_by_attr(target_chunk['handles']['inputs'], 'handle', edge['target_handle']):
                target_chunk['handles']['inputs'].append({
                    'handle': edge['target_handle']
                })
        
        # 针对完全为空的 handle 的 chunk，自动追加默认连接
        for chunk in compiled_chunks:
            if len(chunk['handles']['inputs']) == 0:
                chunk['handles']['inputs'].append(DEFAULT_INPUT_HANDLE.copy())
            if len(chunk['handles']['outputs']) == 0:
                chunk['handles']['outputs'].append(DEFAULT_OUTPUT_HANDLE.copy())

        return {
            'chunks': compiled_chunks,
            'edges': compiled_edges
        }
    
    def create_chunk(self, type = EXECUTOR_TYPE_NORMAL, executor: callable = None, **chunk_desc) -> SchemaChunk:
        """根据 chunk 的描述，创建 chunk 实例"""
        chunk_inst = SchemaChunk(
            workflow_schema=self,
            type=type,
            executor=executor,
            **chunk_desc
        )
        self.append_raw_chunk(chunk_inst.get_raw_schema(use_origin=True))
        return chunk_inst

    def append_raw_chunk(self, chunk: dict):
        """
        添加节点，必须包含 id, type字段，如没有 handles 字段，则会自动追加上默认设置
        """
        # 校验必填字段
        verified_res = validate_dict(chunk, ['id', 'type'])
        if verified_res['status'] == False:
            raise ValueError(f"Missing required key: '{verified_res['key']}'")

        chunk_copy = chunk.copy()
        # 如果没有配置 handles，则自动追加上结构
        if 'handles' not in chunk_copy:
            chunk_copy['handles'] = {
                'inputs': [],
                'outputs': []
            }
        else:
            if 'inputs' not in chunk_copy['handles']:
                chunk_copy['handles']['inputs'] = []
            if 'outputs' not in chunk_copy['handles']:
                chunk_copy['handles']['outputs'] = []
        
        # 检查是否有重名的 chunk
        if has_target_by_attr(self._chunks, 'id', chunk_copy['id']):
            raise ValueError(
                f"The Chunk with the existing id '{chunk_copy['id']}' already exists.")
        self._chunks.append(chunk)
        return self
    
    def append_raw_chunk_list(self, chunks: list):
        """注入多个原生chunk"""
        for chunk in chunks:
            self.append_raw_chunk(chunk)
        return self

    def del_chunk(self, chunk_id):
        """删除节点"""
        self._chunks = [
            chunk for chunk in self._chunks if chunk.get("id") != chunk_id]
        # 清理相关的边(起点/终点为当前节点的边)
        self._edges = [
            edge for edge in self._edges if edge.get("source") != chunk_id and edge.get("target") != chunk_id]
        return self
    
    def get_chunk(self, chunk_id):
        return find_by_attr(self._chunks, 'id', chunk_id)
    
    def get_chunk_by(self, filter: callable):
        return find(self._chunks, filter)

    def connect_chunk(self, source_chunk_id, target_chunk_id, source_handle=DEFAULT_OUTPUT_HANDLE_VALUE, target_handle=DEFAULT_INPUT_HANDLE_VALUE, condition: callable = None, condition_detail: dict = None):
        """
        连接两个节点，分别输入：源节点id、目标节点id、源节点输出句柄（可选，默认为源节点的输出点）、目标节点的输入句柄（可选，默认为目标节点的输入点）
        """
        """
        if source_chunk_id == target_chunk_id:
            raise ValueError(
                f"The starting point and the endpoint have the same value: '{target_chunk_id}'")
        """
        source_chunk = None
        target_chunk = None
        for chunk in self._chunks:
            if chunk['id'] == source_chunk_id:
                source_chunk = chunk
            if chunk['id'] == target_chunk_id:
                target_chunk = chunk

        # 判断节点是否存在
        if source_chunk is None or target_chunk is None:
            raise ValueError('Cannot find target start or end node')

        # 判断句柄是否存在，如果不存在，自动帮其注册
        source_output_handles: list = source_chunk['handles']['outputs']
        if has_target_by_attr(source_output_handles, 'handle', source_handle) == False:
            source_output_handles.append({ 'handle': source_handle })

        target_input_handles: list = target_chunk['handles']['inputs']
        if has_target_by_attr(target_input_handles, 'handle', target_handle) == False:
            target_input_handles.append({'handle': target_handle})
        
        new_edge = {
            'source': source_chunk_id,
            'target': target_chunk_id,
            'source_handle': source_handle or DEFAULT_OUTPUT_HANDLE_VALUE,
            'target_handle': target_handle or DEFAULT_INPUT_HANDLE_VALUE,
            'condition': condition,  # 连通的开关条件，默认直连
            'condition_detail': condition_detail  # 连通的开关条件的类型
        }
        verified_res = self.verify_connection(new_edge)
        if verified_res['status'] == 'error':
            raise ValueError(verified_res['message'])
        elif verified_res['status'] == 'warning':
            self.logger.warn(verified_res["message"])

        self._edges.append(new_edge)
        return self
    
    def verify_connection(self, new_edge: dict):
        """检测连接线的合法性"""
        status_res = { "status": "success", "message": "" }
        is_condition_connection = new_edge.get('condition') is not None
        for edge in self._edges:
            flag = True
            for attr in ['source', 'target', 'source_handle', 'target_handle']:
                if edge.get(attr) != new_edge.get(attr):
                    flag = False
                    break
            # 有相似项，判断是走告警还是走中断，如果是告警，则继续找，看是否有更严重的处理
            if flag:
                source_chunk = self.get_chunk(edge['source'])
                target_chunk = self.get_chunk(edge['target'])
                source_view_name = create_connection_symbol(
                    source_chunk, edge['source_handle'])
                target_view_name = create_connection_symbol(
                    target_chunk, edge['target_handle'])
                # 如果是非条件的情况，直接出报错，然后跳出查找
                if not edge.get('condition') and not is_condition_connection:
                    status_res['status'] = 'error'
                    status_res['message'] = f'An identical connection relationship [ {source_view_name} → {target_view_name} ] already exists.'
                    break
                # 自身为条件连接的情况，直接出 warning，然后跳出查找
                elif is_condition_connection:
                    status_res['status'] = 'warning'
                    status_res['message'] = f'An identical connection relationship [ {source_view_name} → {target_view_name} ] already exists.'
                    break
                # 先写入告警，但继续尝试看是否有报错的情况
                else:
                    status_res['status'] = 'warning'
                    status_res['message'] = f'An identical connection relationship [ {source_view_name} → {target_view_name} ] already exists.'
        return status_res

    
    def get_edge(self, edge_id):
        return find_by_attr(self._edges, 'id', edge_id)
    
    def connect_with_edges(self, edges: list):
        """
        批量添加连接，要求输入为 {"source": str, "target": str, "source_handle"?: str, "target_handle"?: str} 形式构成的数组
        """
        for edge in edges:
            self.connect_chunk(
                edge.get('source'),
                edge.get('target'),
                edge.get('source_handle', DEFAULT_OUTPUT_HANDLE_VALUE),
                edge.get('target_handle', DEFAULT_INPUT_HANDLE_VALUE),
                edge.get('condition') or None,
                edge.get('condition_detail') or None
            )
        return self

    def del_connection(self, source_chunk_id, target_chunk_id, source_handle=DEFAULT_OUTPUT_HANDLE_VALUE, target_handle=DEFAULT_INPUT_HANDLE_VALUE):
        """
        取消连线，分别输入：源节点id、目标节点id、源节点输出句柄（可选，默认为源节点的输出点）、目标节点的输入句柄（可选，默认为目标节点的输入点）
        """
        edges = []
        for edge in self._edges:
            if not (
                edge.get("source") == source_chunk_id and
                edge.get("target") == target_chunk_id and
                (source_handle is None or edge.get('source_handle') == source_handle) and
                (target_handle is None or edge.get('target_handle') == target_handle)
            ):
                edges.append(edge)
        # 有变更时赋值
        if len(edges) != len(self._edges):
            self._edges = edges
        return self
    
    def remove_all_connection(self):
        """移除所有连接关系，但保留注册"""
        self._edges = []
        return self

    def clone(self):
        return Schema(schema_data={
            "chunks": copy.deepcopy(self._chunks),
            "edges": copy.deepcopy(self._edges)
        })



from .utils import FacilityABC
from Agently.utils import RuntimeCtx

class StatusManager(FacilityABC):
    def __init__(self, *, storage: object, plugin_manager: object, settings: object):
        self.storage = storage.table("status_mapping.default")
        self.plugin_manager = plugin_manager

    def set_status_namespace(self, namespace_name: str):
        self.storage = storage.table(f"status_mapping.{ namespace_name }")
        return 

    def set_mappings(self, status_key: str, status_value: str, alias_list: list):
        self.storage.update(f"{status_key}.{status_value}", alias_list).save()
        return self

    def append_mapping(self, status_key: str, status_value: str, alias_name: str, *args, **kwargs):
        self.storage\
            .append(
                f"{status_key}.{status_value}",
                {
                    "alias_name": alias_name,
                    "args": args,
                    "kwargs": kwargs,
                }
            )\
            .save()
        return self

    def get_mapping(self, status_key: str, status_value: str):
        return self.storage.get(f"{status_key}.{status_value}")

def export():
    return ("status_manager", StatusManager)


from ...lib.Store import Store
from ...lib.constants import WORKFLOW_START_DATA_HANDLE_NAME

def start_executor(inputs, store: Store, **sys_info):
    return sys_info['sys_store'].get(WORKFLOW_START_DATA_HANDLE_NAME)



#import nest_asyncio
from .Request import Request
from .Agent import AgentFactory
from .Facility import FacilityManager
from .WebSocket import WebSocketServer, WebSocketClient
from .Workflow import Workflow, Schema as WorkflowSchema
from .AppConnector import AppConnector
from ._global import global_plugin_manager, global_settings, global_storage, global_tool_manager, global_websocket_server
from .utils import *

#nest_asyncio.apply()
def create_agent(*args, **kwargs):
    return AgentFactory().create_agent(*args, **kwargs)

facility = FacilityManager()
lib = facility

set_settings = global_settings.set

def set_compatible(compatible_version: str):
    if isinstance(compatible_version, str):
        compatible_version = int(compatible_version.replace(".", ""))
    set_settings("compatible_version", compatible_version)
    # auto switch settings by compatible version
    if compatible_version <= 3320:
        set_settings("storage_type", "FileStorage")
        set_settings("storage.FileStorage.path", "./.Agently")

def register_plugin(module_name:str, plugin_name: str, plugin: callable):
    global_plugin_manager.register(module_name, plugin_name, plugin)
    facility.refresh_plugins()

def set_plugin_settings(module_name: str, plugin_name: str, key: str, value: any):
    global_plugin_manager.set_settings(f"plugin_settings.{ module_name }.{ plugin_name }", key, value)


# 用于处理中断的逻辑
import time

BREAKING_TYPES = {
    "TIMEOUT": "timeout",
    "CALL_MAXIMUM": "call_maximum"
}


class BreakingHub:
    def __init__(self, breaking_handler, max_execution_limit=5, interval_time = 2, timeout_threshold=10):
        self.max_execution_limit = max_execution_limit
        self.timeout_threshold = timeout_threshold
        self.breaking_handler = breaking_handler
        # 计数间隔时间
        self.interval_time = interval_time
        self.execution_counts = {}
        self.last_break_time = {}

    # 节点执行次数记录器
    def recoder(self, chunk):
        chunk_id = chunk['id']
        if chunk_id not in self.execution_counts:
            self.execution_counts[chunk_id] = 1
            self.last_break_time[chunk_id] = time.time()
        else:
            current_time = time.time()
            if current_time - self.last_break_time[chunk_id] <= self.interval_time:
                self.execution_counts[chunk_id] += 1
            else:
                self.execution_counts[chunk_id] = 1
            self.last_break_time[chunk_id] = current_time

        if self.execution_counts[chunk_id] > self.max_execution_limit:
            self.handle_interrupt(chunk, BREAKING_TYPES["CALL_MAXIMUM"])

    def get_counts(self, chunk):
        chunk_id = chunk['id']
        return self.execution_counts.get(chunk_id, 0)

    def handle_interrupt(self, chunk, type):
        self.breaking_handler(chunk, type)



import inspect
import ast
from .utils import ComponentABC

class Decorator(ComponentABC):
    def __init__(self, agent: object):
        self.agent = agent

    def auto_func(self, func: callable):
        def wrapper(*args, **kwargs):
            # generate input dict
            signature = inspect.signature(func)
            arguments = signature.bind(*args, **kwargs)
            arguments.apply_defaults()
            input_dict = {}
            for param in signature.parameters:
                input_dict.update({ param: arguments.arguments[param] })
            # generate instruction
            instruction = inspect.getdoc(func)
            # generate output dict
            output_dict = signature.return_annotation
            return (
                self.agent
                    .input(input_dict)
                    .instruct(instruction)
                    .output(output_dict)
                    .start()
            )
        return wrapper

    def on_event(self, event: str, *, is_await:bool=False):
        if not event.startswith("response:") and not event.startswith("tool:"):
            event = "response:" + event
        def decorator(func: callable):
            self.agent.add_event_listener(event, func, is_await=is_await)
        return decorator

    def register_tool(self, **tool_info_kwrags):
        def decorator(func: callable):
            # get tool name
            if "tool_name" not in tool_info_kwrags:
                tool_info_kwrags.update({ "tool_name": func.__name__ })
            # get desc
            if "desc" not in tool_info_kwrags:
                tool_info_kwrags.update({ "desc": inspect.getdoc(func) })
            # get args
            if "args" not in tool_info_kwrags:
                func_ast = ast.parse(inspect.getsource(func))
                tool_info_kwrags.update({ "args": {} })
                for node in func_ast.body[0].args.args:
                    if node.arg != "self":
                        tool_info_kwrags["args"].update({
                            node.arg:
                                (node.annotation.dims[0].value, node.annotation.dims[1].value)
                                if isinstance(node.annotation, ast.Tuple)
                                else (type(node.annotation.value).__name__, node.annotation.value)
                        })
            # get func
            tool_info_kwrags.update({ "func": func })
            self.agent.register_tool(**tool_info_kwrags)
            return func
        return decorator

        
    def export(self):
        return {
            "prefix": None,
            "suffix": None,
            "alias": {
                "auto_func": { "func": self.auto_func, "return_value": True },
                "on_event": { "func": self.on_event, "return_value": True },
                "tool": { "func": self.register_tool, "return_value": True },
            },
        }

def export():
    return ("Decorator", Decorator)